from datetime import datetime

# If modifying these scopes, delete the file token.pickle.
SCOPES = [
    "https://www.googleapis.com/auth/calendar.readonly",
    "https://www.googleapis.com/auth/calendar",
]

# Path to task file
task_file_path = "/Users/lukemurray/Dropbox/todo.md"

# the name of the task calendar in google. MUST BE UNIQUE
task_calendar_name = "Markdown Tasks"

# time that you start work every day
work_start_time = datetime.now().replace(hour=8, minute=0, second=0, microsecond=0)

# time that you stop work every day
work_end_time = datetime.now().replace(hour=18, minute=0, second=0, microsecond=0)

# minimum event time in minutes
chunk_size = 15



"""Methods for using the google calendar api."""
import os
import pickle
import typing as t
from datetime import datetime, timedelta

from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from .config import (
    SCOPES,
    task_calendar_name,
)
from .models import (
    CalendarResource,
    EventResource,
    FreeBusyResponse,
    ListCalendar,
    ListCalendarResponse,
)
from .time_helpers import (
    beginning_of_day_utc,
    datetime_to_index,
    end_of_day_utc,
    now_utc,
    parse_rfc_33339,
    total_indices_per_day,
    work_start_utc,
    work_end_utc,
)


def local_file_path(filename):
    return os.path.abspath(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    )


def get_google_api_credentials():
    """Get credentials for accessing the google api."""
    credentials = None
    # The file token.pickle stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    if os.path.exists(local_file_path("token.pickle")):
        with open(local_file_path("token.pickle"), "rb") as token:
            credentials = pickle.load(token)
    # If there are no (valid) credentials available, let the user log in.
    if not credentials or not credentials.valid:
        if credentials and credentials.expired and credentials.refresh_token:
            credentials.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                local_file_path("credentials.json"), SCOPES,
            )
            credentials = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(local_file_path("token.pickle"), "wb") as token:
            pickle.dump(credentials, token)
    return credentials


class CalendarAPI(object):
    def __init__(self):
        super().__init__()
        # service used to access google calendar api
        self.service = build("calendar", "v3", credentials=get_google_api_credentials())
        self._task_cal = None
        self._list_cals = None

    def get_task_calendar(self) -> ListCalendar:
        if self._task_cal is not None:
            return self._task_cal
        # get all existing calendars
        existing_calendars = self.get_calendar_list_calendars()
        # if one already exists return it
        for existing_task_cal in filter(
            lambda c: c["summary"] == task_calendar_name, existing_calendars
        ):
            self._task_cal = existing_task_cal
            return existing_task_cal
        else:
            # otherwise create a new calendar
            calendar: CalendarResource = {
                "summary": task_calendar_name,  # title of the calendar
                "description": "Events generated from my todo.md file",  # calendar description
                "timeZone": "America/New_York",
            }
            created_calendar: CalendarResource = self.service.calendars().insert(
                body=calendar
            ).execute()
            # insert the new calendar in the calendar list
            calendar_list_entry = ListCalendar(id=created_calendar["id"], selected=True)
            created_calendar_list_entry = (
                self.service.calendarList().insert(body=calendar_list_entry).execute()
            )
            self._task_cal = created_calendar_list_entry
            # reset list cals since we added to the calendar list
            self._list_cals = None
            return created_calendar_list_entry

    def get_calendar_list_calendars(self) -> t.List[ListCalendar]:
        """Get all the calendars from the user's calendar list."""
        if self._list_cals is not None:
            return self._list_cals
        all_calendars: t.List[ListCalendar] = []
        page_token = None
        while True:
            calendar_list: ListCalendarResponse = self.service.calendarList().list(
                pageToken=page_token, showHidden=True
            ).execute()
            for calendar_list_entry in calendar_list["items"]:
                all_calendars.append(calendar_list_entry)
            page_token = calendar_list.get("nextPageToken")
            if not page_token:
                break
        self._list_cals = all_calendars
        return all_calendars

    def get_events_from_task_calendar(self) -> t.List[EventResource]:
        all_events: t.List[EventResource] = []
        page_token = None
        while True:
            events = (
                self.service.events()
                .list(
                    calendarId=self.get_task_calendar()["id"],
                    pageToken=page_token,
                    timeZone="UTC",
                    showDeleted=True,
                )
                .execute()
            )
            for event in events["items"]:
                all_events.append(event)
            page_token = events.get("nextPageToken")
            if not page_token:
                break
        return all_events

    def get_visible_calendar_list_calendars(self):
        """Get only those calendars displayed on the user's calendar list."""
        all_calendars = self.get_calendar_list_calendars()
        return [c for c in all_calendars if not c.get("hidden", False)]

    # chunk size in minutes
    def get_busy_array_for_day(self, day_offset: int) -> t.List[bool]:
        # query gcal for free busy results
        timeMin = beginning_of_day_utc(day_offset).isoformat()
        timeMax = end_of_day_utc(day_offset).isoformat()
        free_busy_result: FreeBusyResponse = (
            self.service.freebusy()
            .query(
                body={
                    "timeMin": timeMin,
                    "timeMax": timeMax,
                    "items": [
                        # query all calendars except the task cal
                        {"id": c["id"]}
                        for c in self.get_visible_calendar_list_calendars()
                        if c["summary"] != task_calendar_name
                    ],
                }
            )
            .execute()
        )

        # add free busy results to intervals
        intervals = []
        for c in free_busy_result["calendars"].values():
            for sb in c["busy"]:
                start = parse_rfc_33339(sb["start"])
                end = parse_rfc_33339(sb["end"]) - timedelta(minutes=1)
                intervals.append([start, end])
        # busy from beginning of day to work start time
        intervals.append(
            [
                beginning_of_day_utc(day_offset),
                work_start_utc(day_offset) - timedelta(minutes=1),
            ]
        )
        # busy from work end time to end of day
        intervals.append([work_end_utc(day_offset), end_of_day_utc(day_offset)])

        # busy from beginning of day until now
        if beginning_of_day_utc(day_offset) < now_utc():
            intervals.append([beginning_of_day_utc(day_offset), now_utc()])

        # create an array
        busy = [False] * total_indices_per_day()
        for start, end in intervals:
            # mark times between start and end as busy
            index_start = datetime_to_index(start, day_offset)
            index_end = datetime_to_index(end, day_offset)
            for i in range(index_start, index_end + 1):
                busy[i] = True

        return busy

    def delete_event_from_task_calendar(self, event_id: str):
        self.service.events().delete(
            calendarId=self.get_task_calendar()["id"], eventId=event_id
        ).execute()

    def schedule_task(
        self,
        description: str,
        event_id: str,
        start: datetime,
        end: datetime,
        patch: bool,
    ):
        event_body = {
            "summary": description,
            "start": {"dateTime": start.isoformat()},
            "end": {"dateTime": end.isoformat()},
            "id": event_id,
            "status": "confirmed",
        }
        if patch:
            self.service.events().patch(
                calendarId=self.get_task_calendar()["id"],
                eventId=event_id,
                body=event_body,
            ).execute()
        else:
            self.service.events().patch(
                calendarId=self.get_task_calendar()["id"],
                eventId=event_id,
                body=event_body,
            ).execute()



"""Contains models used throughout the codebase
"""
import typing as t
from typing_extensions import TypedDict
from datetime import datetime, date
from enum import Enum

ETag = t.NewType("ETag", object)


class EventCreator(TypedDict, total=False):
    id: str
    email: str
    displayName: str
    self: bool


class EventOrganizer(TypedDict, total=False):
    id: str
    email: str
    displayName: str
    self: bool


class EventTime(TypedDict, total=False):
    date: str
    dateTime: str
    timeZone: str


class EventAttendee(TypedDict, total=False):
    id: str
    email: str
    displayName: str
    organizer: bool
    self: bool
    resource: bool
    optional: bool
    responseStatus: str
    comment: str
    additionalGuests: int


class EventResource(TypedDict, total=False):
    kind: str
    etag: ETag
    id: str
    status: str
    htmlLink: str
    created: str
    updated: str
    summary: str
    description: str
    location: str
    colorId: str
    creator: EventCreator
    organizer: EventOrganizer
    start: EventTime
    end: EventTime
    endTimeUnspecified: bool
    recurrence: t.Any
    recurringEventId: str
    originalStartTime: EventTime
    transparency: str
    visibility: str
    iCalUID: str
    sequence: str
    attendees: t.List[EventAttendee]
    attendeesOmitted: bool
    extendedProperties: t.Any
    hangoutLink: str
    conferenceData: t.Any
    entryPoints: t.Any
    conferenceSolution: t.Any
    conferenceId: str
    signature: str
    notes: str
    gagdet: t.Any
    reminders: t.Any
    source: t.Any
    attachments: t.Any


class CalendarResource(TypedDict, total=False):
    """Representation of a calendar used in the Calendar's endpoint.

    See https://developers.google.com/calendar/v3/reference/calendars
    """

    kind: str
    id: str
    summary: str
    description: str
    location: str
    timeZone: str
    etag: ETag
    conferenceProperties: t.Any


class ListCalendarDefaultReminders(TypedDict, total=False):
    method: str
    minutes: int


class ListCalendar(TypedDict, total=False):
    kind: str
    etag: ETag
    id: str
    summary: str
    description: str
    location: str
    timeZone: str
    summaryOverride: str
    colorId: str
    backgroundColor: str
    foregroundColor: str
    hidden: bool
    selected: bool
    accessRole: str
    defaultReminders: t.List[ListCalendarDefaultReminders]
    notificationSettings: t.Any
    primary: bool
    deleted: bool
    conferenceProperties: t.Any


class ListCalendarResponse(TypedDict, total=False):
    kind: str
    etag: ETag
    nextPageToken: str
    nextSyncToken: str
    items: t.List[ListCalendar]


class FreeBusyBusyObject(TypedDict, total=False):
    start: str
    end: str


class FreeBusyErrorObject(TypedDict, total=False):
    domain: str
    reason: str


class FreeBusyCalendarValue(TypedDict, total=False):
    errors: t.List[FreeBusyErrorObject]
    busy: t.List[FreeBusyBusyObject]


class FreeBusyResponse(TypedDict, total=False):
    kind: str
    timeMin: str
    timeMax: str
    calendars: t.Dict[str, FreeBusyCalendarValue]


class ParserContext(Enum):
    NONE = 0  # only at the start of the file
    UNK = 1  # unstructured just a string
    TASK = 2  # the default parser context md todolist


class TaskStatus(Enum):
    """The status of a task.

    The comments encode the character used to identify the task in the
    todo field. For example, '- [x]' would be done and '- [c]' would
    be canceled.
    """

    CANCELED = 0  # c
    IN_PROGRESS = 1  # -
    UNSCHEDULED = 2  # ' '
    DONE = 3  # x
    IDEA = 4  # ?
    BLOCKED = 5  # b


class Task(object):
    def __init__(self):
        super().__init__()
        self.status: TaskStatus
        # projects tagged on the task with +project
        self.projects: t.List[str] = []
        # people tagged on the task with +people
        self.people: t.List[str] = []
        # the id of the task (used for syncing with calendar)
        self.id: str
        # the context date of the task i.e. // year-mm-dd
        self.date: datetime.date
        # the task description
        self.description: str
        # the time estimate of the task in minutes
        self.estimate: int
        # the line number of the task in the todo file
        self.line_number: int
        # the children of the task AKA indented sub tasks
        self.children: t.List["Task"] = []
        # day the task is due
        self.due_date: t.Optional[date] = None
        # day the task is scheduled
        self.scheduled_date: t.Optional[date] = None

    def __str__(self):
        return f"{self.description}"



import typing as t
from datetime import datetime, timedelta, timezone
from .config import chunk_size, work_start_time, work_end_time
from dateutil.tz import tzlocal

SECONDS_PER_DAY = 86399


def _apply_day_offset(dt: datetime, day_offset: int) -> datetime:
    """Add or subtract day_offset days from the datetime."""
    return dt + timedelta(days=day_offset)


def beginning_of_day_utc(day_offset: int) -> datetime:
    """Return Local Midnight time of today +/- day_offset days in UTC time."""
    return _apply_day_offset(
        datetime.now().replace(hour=0, minute=0, second=0, microsecond=0), day_offset
    ).astimezone(timezone.utc)


def end_of_day_utc(day_offset: int) -> datetime:
    """Return Local end of day time of today +/- day_offset days in UTC time."""
    return _apply_day_offset(
        datetime.now().replace(hour=23, minute=59, second=59, microsecond=999999),
        day_offset,
    ).astimezone(timezone.utc)


def work_start_utc(day_offset: int) -> datetime:
    """Time that you start work every day."""
    return _apply_day_offset(work_start_time, day_offset).astimezone(timezone.utc)


def work_end_utc(day_offset: int) -> datetime:
    """Time that you end work every day."""
    return _apply_day_offset(work_end_time, day_offset).astimezone(timezone.utc)


def now_utc():
    return datetime.now().astimezone(timezone.utc)


def total_indices_per_day():
    return (24 * 60) // chunk_size


def datetime_to_index(input_date: datetime, day_offset: int) -> int:
    assert input_date.tzinfo == timezone.utc, "expect utc time here"
    time_since_midnight = input_date - beginning_of_day_utc(day_offset)
    seconds_since_midnight = time_since_midnight.total_seconds()
    seconds_per_chunk = chunk_size * 60
    return int(seconds_since_midnight // seconds_per_chunk)


def index_to_timestamp(index: int, day_offset: int):
    seconds_per_chunk = chunk_size * 60
    seconds_since_midnight = index * seconds_per_chunk
    return beginning_of_day_utc(day_offset).timestamp() + seconds_since_midnight


def start_index_to_datetime(index: int, day_offset: int):
    unix_time = index_to_timestamp(index, day_offset)
    dt = datetime.fromtimestamp(unix_time, timezone.utc)
    dt -= timedelta(
        minutes=dt.minute % chunk_size, seconds=dt.second, microseconds=dt.microsecond
    )
    return dt


def end_index_to_datetime(index: int, day_offset: int):
    unix_time = index_to_timestamp(index, day_offset)
    dt = datetime.fromtimestamp(unix_time, timezone.utc)
    dt -= timedelta(
        minutes=dt.minute % chunk_size, seconds=dt.second, microseconds=dt.microsecond
    )
    dt -= timedelta(seconds=1)
    return dt


def utc_to_local(utc_datetime: datetime):
    return utc_datetime.astimezone(tzlocal())


def parse_rfc_33339(rfc3339: str) -> datetime:
    return datetime.strptime(rfc3339, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)


def debug_busy_array(busy_array: t.List[bool], day_offset: int):
    print("\n\n---------- BUSY ARRAY ----------\n\n")
    for index, busy in enumerate(busy_array):
        start = utc_to_local(start_index_to_datetime(index, day_offset))
        end = utc_to_local(end_index_to_datetime(index + 1, day_offset))
        print(
            f'{start.strftime("%H:%M||%m:%d")} - {end.strftime("%H:%M||%m:%d")} : {"busy" if busy else "free"}'
        )




"""Console commands for the scheduler."""
from __future__ import print_function
from .config import task_file_path
from .task_parser import TaskParser
from .scheduler import schedule_tasks


def run():
    """Run the scheduler.

    This command is exported as the script for the package.
    """
    tasks = TaskParser().get_tasks(task_file_path)
    schedule_tasks(tasks)



import re
import typing as t
from datetime import datetime

from .models import ParserContext, Task, TaskStatus


def _get_status_from_char(char: str) -> TaskStatus:
    """Convert the character in the [ ] for a task into a TaskStatus

    Args:
        char (str): the character in the brackets of a task

    Raises:
        Exception: If the character is not mapped to a TaskStatus

    Returns:
        TaskStatus: The status of the task.
    """
    if char == "c":
        return TaskStatus.CANCELED
    elif char == "b":
        return TaskStatus.BLOCKED
    elif char == " ":
        return TaskStatus.UNSCHEDULED
    elif char == "x":
        return TaskStatus.DONE
    elif char == "?":
        return TaskStatus.IDEA
    elif char == "-":
        return TaskStatus.IN_PROGRESS
    else:
        raise Exception("unknown status character {char}".format(char=char))


# the parser
class TaskParser(object):
    def __init__(self):
        super().__init__()
        self.parser_context: ParserContext = ParserContext.NONE
        self.date_context: datetime.date
        self.tasks: t.List[Task] = []
        self.task_ids = set()
        self.indent_type = None
        self.indent_level_to_task: t.Dict[Task] = dict()

    def _parse_line(self, line: str, line_number: int):
        # ignore white space lines
        if re.fullmatch(r"\s+", line):
            pass
        # context lines start with '//'
        elif line.startswith("//"):
            try:
                new_date_context = datetime.strptime(
                    line[2:].strip(), "%Y-%m-%d"
                ).date()
                self.date_context = new_date_context
                self.parser_context = ParserContext.TASK
            # TODO(lukemurray): fail if unknown context
            except ValueError:
                self.parser_context = ParserContext.NONE
        else:
            # parse non blank and non context lines based on the current context
            if self.parser_context == ParserContext.TASK:
                self._parse_task(line, line_number)
            elif self.parser_context == ParserContext.NONE:
                pass
            else:
                raise Exception(
                    "no parsing handled for parse context {context}".format(context=self.parser_context)
                )

    def _check_indent(self, indent: str, line_number: str):
        if len(indent) != 0:
            all_tabs = all(c == "\t" for c in indent)
            all_spaces = all(c == " " for c in indent)
            if all_tabs or all_spaces:
                expected_indent_type = "tabs" if all_tabs else "spaces"
                if self.indent_type is None:
                    self.indent_type = expected_indent_type
                elif self.indent_type != expected_indent_type:
                    raise Exception(
                        "Prev indent character {indent_type} does not \
                            match indent character {expected_indent_type} on \
                                line {line_number}".format(indent_type=self.indent_type, expected_indent_type=expected_indent_type, line_number=line_number)
                    )
            else:
                raise Exception(
                    "Please do not mix tabs and spaces in your task indent (line: {line_number})".format(line_number=line_number)
                )

    def _parse_task(self, line: str, line_number: int):
        new_task = Task()
        # make sure the line starts with spaces followed by - [ ] there can be any character in the brackets
        task_status_regex = r"(\s*)\- \[(.?)\]"
        task_status_match = re.match(task_status_regex, line)
        if not task_status_match:
            raise Exception("(line {line_number}) todos must start with - [ ]".format(line_number=line_number))
        else:
            assert task_status_match is not None
            indent: str = task_status_match.group(1)
            self._check_indent(indent, line_number)
            status_char = task_status_match.group(2)
            new_task.status = _get_status_from_char(status_char)

        # add people to task
        person_regex = r"@(\w+)"
        people = re.finditer(person_regex, line)
        for person_match in people:
            new_task.people.append(person_match.group(1))

        # add project to task
        project_regex = r"\+(\w+)"
        projects = re.finditer(project_regex, line)
        for project_match in projects:
            new_task.projects.append(project_match.group(1))

        # get the tags as a dict
        tag_regex = r"(\w+):([\w|\-]+)"
        tags = re.finditer(tag_regex, line)
        tag_dict: t.Dict[str, str] = dict()
        for tag_match in tags:
            tag_dict[tag_match.group(1)] = tag_match.group(2)
        if "id" not in tag_dict:
            raise Exception("(line {line_number}) task is missing an id".format(line_number=line_number))
        # add the id to the task
        new_task.id = tag_dict["id"]
        if new_task.id in self.task_ids:
            raise Exception("duplicate ids")
        self.task_ids.add(new_task.id)

        # TODO(lukemurray): maybe depreacted by the date tag
        # add the date to the task
        new_task.date = self.date_context

        # parse the estimate tag
        est = tag_dict.get("est", "30m")
        est_regex = r"(\d+)(\w+)"
        est_in_minutes = 30
        est_match = re.fullmatch(est_regex, est)
        if est_match is not None:
            measure = est_match.group(1)
            unit = est_match.group(2)
            if unit == "m":
                est_in_minutes = int(measure)
            elif unit == "h":
                est_in_minutes = 60 * int(measure)
            else:
                raise Exception("unknown unit {unit}".format(unit=unit))

        # add the estimate to the task
        new_task.estimate = est_in_minutes

        # description from removing status and tags
        new_task.description = (
            re.sub(tag_regex, "", re.sub(task_status_regex, "", line))
            .replace("\n", "")
            .replace("\t", "")
            .strip()
        )

        new_task.line_number = line_number

        # manage indent dependencies
        # indent level is like the number of spaces and can be 1, 2, 3, 4 etc.
        # semantic indent level is just based on hiearchy and ignored cardinality
        # indirection to ignore spacing issues and provide nice indexing
        indent_level = len(indent)
        self.indent_level_to_task[indent_level] = new_task
        all_indent_levels = sorted(list(self.indent_level_to_task.keys()))
        semantic_indent_level = all_indent_levels.index(indent_level)
        if semantic_indent_level != 0:
            parent_task = self.indent_level_to_task[
                all_indent_levels[semantic_indent_level - 1]
            ]
            parent_task.children.append(new_task)

        # due date
        if "due" in tag_dict:
            due_date = datetime.strptime(tag_dict["due"], "%Y-%m-%d").date()
            new_task.due_date = due_date

        self.tasks.append(new_task)

    def parse(self, file: str):
        with open(file, "r") as f:
            for line_number, line in enumerate(f):
                self._parse_line(line, line_number)

    def get_tasks(self, file: str):
        self.parse(file)
        return self.tasks



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reschedule']

package_data = \
{'': ['*']}

install_requires = \
['google-api-python-client>=1.7.11,<2.0.0',
 'google-auth-oauthlib>=0.4.1,<0.5.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'typing-extensions>=3.7.4,<4.0.0']

entry_points = \
{'console_scripts': ['reschedule = reschedule.console:run']}

setup_kwargs = {
    'name': 'reschedule',
    'version': '0.1.0',
    'description': 'Parser and scheduler for syncing a markdown(ish) todo list with google calendar.',
    'long_description': '# TODO MD Scheduler\n\n## Goal\n\nI have a markdown file of todos which I want to automatically schedule on my google calendar.\n\n```markdown\n- [ ] task name @related_person +task_project est:10m id:1f2a4b\n```\n\nEach task is assumed to have the following properties\n\n- **id**: a unique id\n- **description**: a description of the task\n- **estimate**: a time estimate written as `5m` or `2h` for 5 minutes and 2 hours respectively.\n\nThis program divides a day into chunks of size 15 minutes and schedules tasks during those chunks. The tasks are added to a unique calendar in Google Calendar.\n\n## Design\n\n`task_parser.py` parses the todo file and creates a list of tagged `task` objects.\n\n`calendar_api.py` is in charge of getting calendar information.\n',
    'author': 'Luke Murray',
    'author_email': 'lukepigeonmail@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lukesmurray/reschedule',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



import math
import sys
import typing as t

from .calendar_api import CalendarAPI
from .config import chunk_size
from .models import TaskStatus
from .task_parser import Task
from .time_helpers import (
    end_index_to_datetime,
    now_utc,
    parse_rfc_33339,
    start_index_to_datetime,
    utc_to_local,
)


def schedule_tasks(parse_tasks: t.List[Task]):
    cal_api = CalendarAPI()
    cal_tasks = cal_api.get_events_from_task_calendar()
    cal_task_ids = {c["id"] for c in cal_tasks}
    # tasks which are already deleted
    cancelled_cal_task_ids = {c["id"] for c in cal_tasks if c["status"] == "cancelled"}
    parse_task_ids = {c.id for c in parse_tasks}
    # delete tasks which are in calendar but not in parse tasks
    cal_task_ids_to_delete = cal_task_ids.difference(parse_task_ids).difference(
        cancelled_cal_task_ids
    )
    for event_id in cal_task_ids_to_delete:
        cal_api.delete_event_from_task_calendar(event_id)
    cancelled_cal_task_ids = cancelled_cal_task_ids.union(cancelled_cal_task_ids)

    # ids of tasks which are not finished
    unfinished_task_status_set = set([TaskStatus.UNSCHEDULED, TaskStatus.IN_PROGRESS])
    parse_task_ids_not_finished = set(
        [t.id for t in parse_tasks if t.status in unfinished_task_status_set]
    )

    # tasks in the future on the calendar
    future_cal_task_ids = {
        c["id"] for c in cal_tasks if parse_rfc_33339(c["end"]["dateTime"]) > now_utc()
    }

    # delete tasks in the future that are finished
    cal_task_ids_to_delete = future_cal_task_ids.difference(
        parse_task_ids_not_finished
    ).difference(
        cancelled_cal_task_ids
    )  # make an exception for already deleted tasks
    for event_id in cal_task_ids_to_delete:
        cal_api.delete_event_from_task_calendar(event_id)
    for tid in cal_task_ids_to_delete:
        cancelled_cal_task_ids.add(tid)

    # reschedule tasks which are not finished but are on the calendar
    task_ids_to_reschedule = parse_task_ids_not_finished.intersection(cal_task_ids)
    # schedule tasks ids which are not finished and are not on the calendar
    task_ids_to_schedule = parse_task_ids_not_finished.difference(cal_task_ids)
    task_id_dict = {t.id: t for t in parse_tasks}
    # list of all tasks to add to the calendar
    all_tasks_to_add_to_cal = [
        task_id_dict[tid] for tid in task_ids_to_reschedule.union(task_ids_to_schedule)
    ]
    # sort list by line number
    all_tasks_to_add_to_cal.sort(key=lambda t: t.line_number)
    # create a dict of id to task sorted by line number
    sorted_task_dict = {t.id: task_id_dict[t.id] for t in all_tasks_to_add_to_cal}

    # the actual scheduling
    day_offset = 0
    busy_array = cal_api.get_busy_array_for_day(day_offset)

    while len(sorted_task_dict):
        try:
            # get the first possible free time
            start_free_index = busy_array.index(False)
        except ValueError:
            # no free time remaining so go to next day
            day_offset += 1
            busy_array = cal_api.get_busy_array_for_day(day_offset)
            continue

        try:
            # get the next busy time after the free time
            next_busy_index = busy_array.index(True, start_free_index)
        except ValueError:
            # no more busy times so we are free til the end of the day
            next_busy_index = len(busy_array)

        # free time in minutes
        free_time = (next_busy_index - start_free_index) * chunk_size
        task_ids_which_fit = filter(
            lambda tid: sorted_task_dict[tid].estimate <= free_time, sorted_task_dict
        )

        for fitting_task_id in task_ids_which_fit:
            fitting_task = sorted_task_dict[fitting_task_id]
            # schedule the task
            # get the start and end time
            task_end_index = start_free_index + math.ceil(
                fitting_task.estimate / chunk_size
            )
            start_time = start_index_to_datetime(start_free_index, day_offset)
            end_time = end_index_to_datetime(task_end_index, day_offset)
            # schedule the task
            cal_api.schedule_task(
                fitting_task.description,
                fitting_task.id,
                start_time,
                end_time,
                fitting_task.id in task_ids_to_reschedule,
            )
            # remove task from sorted task dict
            sorted_task_dict.pop(fitting_task.id)
            fitting_task.scheduled_date = utc_to_local(start_time).date()
            # mark allotted time as busy
            for i in range(start_free_index, task_end_index):
                busy_array[i] = True
            break
        else:
            # no task could be scheduled mark entire time as busy
            for i in range(start_free_index, next_busy_index):
                busy_array[i] = True

    error_mesage = ""
    for task in all_tasks_to_add_to_cal:
        if task.due_date is not None and task.scheduled_date is not None:
            if task.due_date <= task.scheduled_date:
                error_mesage += (
                    f'ERROR: failed to schedule the task "{task}" before due date.\n'
                )
                error_mesage += f"\tscheduled for {task.scheduled_date:%Y-%m-%d} and due on {task.due_date:%Y-%m-%d}\n"
                error_mesage += "FIX: break up long tasks into smaller ones or change priority by placing task earlier in file.\n"

    if error_mesage:
        sys.exit(error_mesage)
